@@ -20,6 +20,7 @@
 #include "ce_actions_mt.h"
 #include "ce_comm.h"
 #include "ce_trace.h"
+#include "ce_appl_env_list.h"
 
 #include <stdio.h>
 /**************************************************************/
@@ -32,20 +33,22 @@ static struct global {
     int initialized;
     ce_lck_t *mutex;
 
-    ce_mt_queue_t *q;
-    ce_mt_queue_t *internal_q;
-
     CE_SOCKET sock ;  /* Used to communicate with the Ensemble thread */
     int req;    /* Have we requested attention already? */
     int checking ;
+    ce_mt_queue_t *q;
+    ce_mt_queue_t *internal_q;
     ce_sema_t *wait_AddSockRecv;
     ce_sema_t *wait_RmvSockRecv;
     struct sockaddr_in to;
+    ce_pool_t *mt_pool;                /* The action pool used by MT */
+    ce_pool_t *st_pool;                /* The action pool used by ST */
+    ce_appl_env_list_t *appl_env_list;
 } g;
 
 /* Initialize the global state
  */ 
-static void
+static void 
 init_g (void)
 {
     TRACE("init_g(");
@@ -57,9 +60,13 @@ init_g (void)
     g.checking = 0;
     g.wait_AddSockRecv = ce_sema_Create(1);
     g.wait_RmvSockRecv = ce_sema_Create(1);
-
+    g.mt_pool = ce_pool_create();
+    g.st_pool = ce_st_get_allocation_pool();
+    g.appl_env_list = ce_appl_env_list_create();
+    
     ce_internal_sock(&g.sock, &g.to);
 
+	
     TRACE(")");
 }
 
@@ -67,6 +74,9 @@ init_g (void)
 
 typedef struct {
     ce_env_t user_env;
+    ce_appl_intf_t *c_appl;
+    ce_queue_t *mt_aq;               /* A queue of pending actions */
+    int req_heartbeat;   /* Should we request a callback? */
     ce_appl_exit_t exit;
     ce_appl_install_t install;
     ce_appl_flow_block_t flow_block;
@@ -89,6 +99,8 @@ create_mt_env(ce_env_t env,
     mt_env_t *e;
 
     e = (mt_env_t*) ce_malloc(sizeof(mt_env_t));
+    memset(e, 0, sizeof(mt_env_t));
+    e->mt_aq = ce_queue_create();
     e->user_env = env;
     e->exit = exit;
     e->install = install;
@@ -97,12 +109,14 @@ create_mt_env(ce_env_t env,
     e->cast = cast;
     e->send = send;
     e->heartbeat = heartbeat;
+
     return e;
 }
 
 static void
 destroy_mt_env(mt_env_t *e)
 {
+    ce_queue_free(g.mt_pool, e->mt_aq);
     free(e);
 }
 
@@ -111,95 +125,87 @@ destroy_mt_env(mt_env_t *e)
 /* Takes the ens_mutex lock, and performs all pending Ensemble
  * operations. 
  */
-static void
-check_pending(void)
+static void check_pending(void)
 {
     int i;
     ce_mt_action_t *a;
+    int total_num_actions = 0;
     
     /* Prevent re-entry by the ens_thread.
      */
     if (g.checking) return;
     g.checking = 1;
 
+    if (rand() % 100 == 0) {
+	TRACE_GEN(("mt_pool free=%d\n", g.mt_pool->num_free));
+	TRACE_GEN(("st_pool free=%d\n", g.st_pool->num_free));
+    }
 //    TRACE("check_pending(");
     ce_lck_Lock(g.mutex);
-    if (g.q->len > 0) {
+    if (g.q->len > 0)
 	ce_mt_queue_copy_and_clear(g.q, g.internal_q);
-	ce_lck_Unlock(g.mutex);
-	g.req = 0;
-	
-	/* Dispatch all pending events. 
-	 */
-	for(i=0; i<g.internal_q->len; i++){
-	    a = (ce_mt_action_t *) &(g.internal_q->arr[i]);
 
-	    switch (a->type){
-	    case MT_CAST:
-//		TRACE("MT_CAST");
-		ce_st_Cast(a->c_appl, a->u.cast.num, a->u.cast.iovl);
-		break;
-	    case MT_SEND:
-//		TRACE("MT_SEND");
-		ce_st_Send(a->c_appl, a->u.send.num_dests, a->u.send.dests,
-			      a->u.send.num, a->u.send.iovl);
-		break;
-	    case MT_SEND1:
-//		TRACE("MT_SEND1");
-		ce_st_Send1(a->c_appl, a->u.send1.dest, a->u.send1.num, a->u.send1.iovl);
-		break;
-	    case MT_LEAVE:
-//		TRACE("MT_LEAVE");
-		ce_st_Leave(a->c_appl);
-		break;
-	    case MT_PROMPT:
-//		TRACE("MT_PROMPT");
-		ce_st_Prompt(a->c_appl);
-		break;
-	    case MT_SUSPECT:
-//		TRACE("MT_SUSPECT");
-		ce_st_Suspect(a->c_appl, a->u.suspect.num, a->u.suspect.members);
-		break;
-	    case MT_XFERDONE:
-//		TRACE("MT_XFERDONE");
-		ce_st_XferDone(a->c_appl);
-		break;
-	    case MT_REKEY:
-//		TRACE("MT_REKEY");
-		ce_st_Rekey(a->c_appl);
-		break;
-	    case MT_PROTOCOL:
-//		TRACE("MT_PROTOCOL");
-		ce_st_ChangeProtocol(a->c_appl, a->u.proto);
-		break;
-	    case MT_PROPERTIES:
-//		TRACE("MT_PROPERITES");
-		ce_st_ChangeProperties(a->c_appl, a->u.properties);
-		break;
-		
-	    case MT_JOIN:
-//		TRACE("MT_JOIN");
-		ce_st_Join(a->u.join.ops, a->c_appl);
-		break;
-	    case MT_ADDSOCKRECV:
-		ce_st_AddSockRecv(a->u.add_sock.socket,
-				     a->u.add_sock.handler, a->u.add_sock.env);
-		ce_sema_Inc(g.wait_AddSockRecv);
-		break;
-	    case MT_RMVSOCKRECV:
-		ce_st_RmvSockRecv(a->u.rmv_sock.socket);
-		ce_sema_Inc(g.wait_RmvSockRecv);
-		break;
-	    }
-	}
-	/* cleanup
-	 */
-	ce_mt_queue_clear (g.internal_q);
-    } else {
-	ce_lck_Unlock(g.mutex);
-	g.req = 0;
+    for(i=0; i<g.appl_env_list->len; i++) {
+	mt_env_t *env;
+	
+	if (g.appl_env_list->arr[i] == 0) continue;
+	
+	env = (mt_env_t*) g.appl_env_list->arr[i];
+	if (ce_queue_length(env->mt_aq) == 0) continue;
+
+	total_num_actions += ce_queue_length(env->mt_aq);
+	ce_queue_pour(env->mt_aq, env->c_appl->aq);
+	env->req_heartbeat = 1;
     }
+
+    /* Replenish the MT pool from the ST pool
+     */
+    ce_pool_move(g.st_pool, g.mt_pool, total_num_actions);
     
+    ce_lck_Unlock(g.mutex);
+    g.req = 0;
+    
+    /* These parts do not require the lock. They can also take a long time, since they call
+     * Ensemble synchronously.
+     */
+    for(i=0; i<g.appl_env_list->len; i++) {
+	mt_env_t *env;
+	
+	if (g.appl_env_list->arr[i] == 0) continue;
+	
+	env = (mt_env_t*) g.appl_env_list->arr[i];
+	if (0 == env->req_heartbeat) continue;
+	
+	ce_st_check_heartbeat(env->c_appl);
+	env->req_heartbeat = 0;
+    }
+
+    /* Dispatch all pending events. 
+     */
+    for(i=0; i<g.internal_q->len; i++){
+	a = (ce_mt_action_t *) &(g.internal_q->arr[i]);
+	
+	switch (a->type){
+	case MT_JOIN:
+//		TRACE("MT_JOIN");
+	    ce_st_Join(a->u.join.ops, a->u.join.c_appl);
+	    ce_jops_free(a->u.join.ops);
+	    break;
+	case MT_ADDSOCKRECV:
+	    ce_st_AddSockRecv(a->u.add_sock.socket,
+			      a->u.add_sock.handler, a->u.add_sock.env);
+	    ce_sema_Inc(g.wait_AddSockRecv);
+	    break;
+	case MT_RMVSOCKRECV:
+	    ce_st_RmvSockRecv(a->u.rmv_sock.socket);
+	    ce_sema_Inc(g.wait_RmvSockRecv);
+	    break;
+	}
+    }
+    /* cleanup
+     */
+    ce_mt_queue_clear (g.internal_q);
+
     g.checking = 0;
 //    TRACE(")");
 }
@@ -211,19 +217,10 @@ mt_exit(ce_env_t env)
 {
     mt_env_t *e = (mt_env_t*) env;
     e->exit(e->user_env);
+    ce_appl_env_list_remove(g.appl_env_list, (void*)e);
     destroy_mt_env(e);
 }
 
-/*
-#define REQ  { \
-    ce_lck_Lock(g.mutex); \
-    g.req = 1; \
-    ce_lck_Unlock(g.mutex);\
-}
-*/
-
-#define REQ g.req = 1; 
-
 /* Call the application install handler, and then send an asynchronous
  * block_ok event to Ensemble.
  */
@@ -231,7 +228,7 @@ static void
 mt_install(ce_env_t env, ce_local_state_t *ls, ce_view_state_t *vs)
 {
     mt_env_t *e = (mt_env_t*)env;
-    REQ;
+    g.req = 1;
     e->install(e->user_env, ls, vs);
     check_pending();
 }
@@ -240,7 +237,7 @@ static void
 mt_flow_block(ce_env_t env, ce_rank_t origin, ce_bool_t onoff)
 {
     mt_env_t *e = (mt_env_t*)env;
-    REQ;
+    g.req = 1;
     TRACE("mt_flow_block");
     e->flow_block(e->user_env, origin, onoff);
     check_pending();
@@ -250,7 +247,7 @@ static void
 mt_block(ce_env_t env)
 {
     mt_env_t *e = (mt_env_t*)env;
-    REQ;
+    g.req = 1;
     e->block(e->user_env);
     check_pending();
 }
@@ -259,7 +256,7 @@ static void
 mt_cast(ce_env_t env, ce_rank_t origin, int len, ce_iovec_array_t iovl)
 {
     mt_env_t *e = (mt_env_t*)env;
-    REQ;
+    g.req = 1;
     e->cast(e->user_env, origin, len, iovl);
     check_pending();
 }
@@ -268,7 +265,7 @@ static void
 mt_send(ce_env_t env, ce_rank_t origin, int len, ce_iovec_array_t iovl)
 {
     mt_env_t *e = (mt_env_t*)env;
-    REQ;
+    g.req = 1;
     e->send(e->user_env, origin, len, iovl);
     check_pending();
 }
@@ -277,7 +274,7 @@ static void
 mt_heartbeat(ce_env_t env, ce_time_t time)
 {
     mt_env_t *e = (mt_env_t*)env;
-    REQ;
+    g.req = 1;
     e->heartbeat(e->user_env, time);
     check_pending();
 }
@@ -288,22 +285,26 @@ mt_heartbeat(ce_env_t env, ce_time_t time)
  */
 ce_appl_intf_t*
 ce_mt_create_intf(ce_env_t env, 
-	       ce_appl_exit_t exit,
-	       ce_appl_install_t install,
-	       ce_appl_flow_block_t flow_block,
-	       ce_appl_block_t block,
-	       ce_appl_receive_cast_t cast,
-	       ce_appl_receive_send_t send,
-	       ce_appl_heartbeat_t heartbeat
+		  ce_appl_exit_t exit,
+		  ce_appl_install_t install,
+		  ce_appl_flow_block_t flow_block,
+		  ce_appl_block_t block,
+		  ce_appl_receive_cast_t cast,
+		  ce_appl_receive_send_t send,
+		  ce_appl_heartbeat_t heartbeat
     ){
     mt_env_t *e;
+    ce_appl_intf_t *c_appl;
     
     TRACE("ce_create_intf");
     e = create_mt_env(env, exit, install, flow_block, block, cast, send, heartbeat);
-    return ce_st_create_intf((void*)e, mt_exit,
-				mt_install, mt_flow_block,
-				mt_block, mt_cast,
-				mt_send, mt_heartbeat);
+    ce_appl_env_list_add(g.appl_env_list, (void*)e);
+    c_appl = ce_st_create_intf((void*)e, mt_exit,
+			       mt_install, mt_flow_block,
+			       mt_block, mt_cast,
+			       mt_send, mt_heartbeat);
+    e->c_appl = c_appl;
+    return c_appl;
 }
 
 /* Send one byte on the socket to Ensemble. This call is unsafe, must
@@ -314,7 +315,7 @@ void
 req_attention(void)
 {
     int result = 0;
-    char buf[2] = "a";
+    char *buf = "a";
 
     if (g.req == 0) {
 	g.req = 1;
@@ -356,9 +357,7 @@ sock_handler(void *dummy)
      /* Receive one byte, this should be thread-safe.
       */
     
-//    ce_lck_Lock(g.mutex);
     res = recv(g.sock, (char*)&buf, 4, 0);
-//    ce_lck_Unlock(g.mutex);
     check_pending();
     TRACE("]");
 }
@@ -383,9 +382,12 @@ void
 ce_mt_Init(int argc,char **argv)
 {
     TRACE("ce_Init(");
-    init_g();
     ce_st_Init(argc, argv);
 
+    /* MT initalization is dependent upon ST initialization. 
+     */
+    init_g();
+
     /* This must be performed exactly once.
      */
     TRACE("ce_Main_loop(");
@@ -423,23 +425,25 @@ ce_mt_Join(ce_jops_t *ops, ce_appl_intf_t *c_appl)
     TRACE(")");
 }
 
-void
-ce_mt_Leave(ce_appl_intf_t *c_appl)
+void ce_mt_Leave(ce_appl_intf_t *c_appl)
 {
+    mt_env_t *env = c_appl->env;
+    
     TRACE("Leave("); 
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_leave(g.q, c_appl);
+    ce_action_leave(g.mt_pool, env->mt_aq);
     ce_lck_Unlock(g.mutex);
     req_attention();
     TRACE(")");
 }
 
-void
-ce_mt_Cast(ce_appl_intf_t *c_appl, int num, ce_iovec_array_t iovl)
+void ce_mt_Cast(ce_appl_intf_t *c_appl, int num, ce_iovec_array_t iovl)
 {
+    mt_env_t *env = c_appl->env;
+
     TRACE("Cast(");
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_cast(g.q, c_appl, num, iovl);
+    ce_action_cast (g.mt_pool, env->mt_aq, num, iovl);
     ce_lck_Unlock(g.mutex);
     req_attention();
     TRACE(")");
@@ -447,34 +451,40 @@ ce_mt_Cast(ce_appl_intf_t *c_appl, int num, ce_iovec_array_t iovl)
 
 void
 ce_mt_Send(ce_appl_intf_t *c_appl, int num_dests, ce_rank_array_t dests,
-	int num, ce_iovec_array_t iovl)
+	   int num, ce_iovec_array_t iovl)
 {
-   TRACE("Send(");
+    mt_env_t *env = c_appl->env;
+
+    TRACE("Send(");
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_send(g.q, c_appl, num_dests, dests, num, iovl);
+    ce_action_send(g.mt_pool, env->mt_aq, num_dests, dests, num, iovl);
     ce_lck_Unlock(g.mutex);
     req_attention();
-   TRACE(")");
+    TRACE(")");
 }
 
 
 void
 ce_mt_Send1(ce_appl_intf_t *c_appl, ce_rank_t dest, int num,
-	 ce_iovec_array_t iovl)
+	    ce_iovec_array_t iovl)
 {
-   TRACE("Send1(");
+    mt_env_t *env = c_appl->env;
+
+    TRACE("Send1(");
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_send1(g.q, c_appl, dest, num, iovl);
+    ce_action_send1(g.mt_pool, env->mt_aq, dest, num, iovl);
     ce_lck_Unlock(g.mutex);
     req_attention();
-   TRACE(")");
+    TRACE(")");
 }
 
 void
 ce_mt_Prompt(ce_appl_intf_t *c_appl)
 {
+    mt_env_t *env = c_appl->env;
+
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_prompt(g.q, c_appl);
+    ce_action_prompt(g.mt_pool, env->mt_aq);
     ce_lck_Unlock(g.mutex);
     req_attention();
 }
@@ -482,8 +492,10 @@ ce_mt_Prompt(ce_appl_intf_t *c_appl)
 void
 ce_mt_Suspect(ce_appl_intf_t *c_appl, int num, ce_rank_array_t suspects)
 {
+    mt_env_t *env = c_appl->env;
+
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_suspect(g.q, c_appl, num, suspects);
+    ce_action_suspect(g.mt_pool, env->mt_aq, num, suspects);
     ce_lck_Unlock(g.mutex);
     req_attention();
 }
@@ -491,9 +503,11 @@ ce_mt_Suspect(ce_appl_intf_t *c_appl, int num, ce_rank_array_t suspects)
 void
 ce_mt_XferDone(ce_appl_intf_t *c_appl)
 {
+    mt_env_t *env = c_appl->env;
+
     TRACE("ce_XferDone");
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_xfer_done(g.q, c_appl);
+    ce_action_xfer_done(g.mt_pool, env->mt_aq);
     ce_lck_Unlock(g.mutex);
     req_attention();
 }
@@ -501,10 +515,11 @@ ce_mt_XferDone(ce_appl_intf_t *c_appl)
 void
 ce_mt_Rekey(ce_appl_intf_t *c_appl)
 {
+    mt_env_t *env = c_appl->env;
 
     TRACE("ce_Rekey");
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_rekey(g.q, c_appl);
+    ce_action_rekey(g.mt_pool, env->mt_aq);
     ce_lck_Unlock(g.mutex);
     req_attention();
 }
@@ -512,8 +527,10 @@ ce_mt_Rekey(ce_appl_intf_t *c_appl)
 void
 ce_mt_ChangeProtocol(ce_appl_intf_t *c_appl, char *protocol_name)
 {
+    mt_env_t *env = c_appl->env;
+
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_protocol(g.q, c_appl, protocol_name);
+    ce_action_protocol(g.mt_pool, env->mt_aq, protocol_name);
     ce_lck_Unlock(g.mutex);
     req_attention();
 }
@@ -521,8 +538,10 @@ ce_mt_ChangeProtocol(ce_appl_intf_t *c_appl, char *protocol_name)
 void
 ce_mt_ChangeProperties(ce_appl_intf_t *c_appl,  char *properties)
 {
+    mt_env_t *env = c_appl->env;
+
     ce_lck_Lock(g.mutex);
-    ce_mt_appl_properties(g.q, c_appl, properties);
+    ce_action_properties(g.mt_pool, env->mt_aq, properties);
     ce_lck_Unlock(g.mutex);
     req_attention();
 }
