@@ -65,7 +65,7 @@ ce_thread_Create(void (*routine)(void*), void *arg, int stacksize)
 
     TRACE("starting thread(\n");
     if (CreateThread(NULL, stacksize, (LPTHREAD_START_ROUTINE)routine, arg, 0, &dwThreadID) == NULL)
-	ce_panic("hot_thread_Create: t_create");
+	ce_panic("ce_thread_Create: t_create");
     TRACE(")\n");
 }
 
@@ -78,11 +78,11 @@ ce_sema_Create(int initial_value)
     
     sp = (ce_sema_t*) malloc(sizeof(ce_sema_t));
     if (sp == NULL)
-	ce_panic("hot_sema_Create:  malloc returned NULL");
+	ce_panic("ce_sema_Create:  malloc returned NULL");
     
     if ((sp->sema = CreateSemaphore(NULL, initial_value, 0x7fffffff, NULL)) == NULL) {
         printf("gle %d\n",GetLastError());
-        ce_panic("hot_sema_Create: failed to create semaphore");
+        ce_panic("ce_sema_Create: failed to create semaphore");
     }
     
     return sp;
@@ -204,7 +204,7 @@ ce_thread_Create(void (*routine)(void*),
 #endif
     
     if (pthread_create(&td, &attr, (ce_thread_routine) routine, arg))
-	ce_panic("hot_thread_Create: pthread_create");
+	ce_panic("ce_thread_Create: pthread_create");
     pthread_attr_destroy(&attr);
 }
 
@@ -219,7 +219,7 @@ ce_sema_Create(int initial_value)
     sp = (ce_sema_t*) ce_malloc(sizeof(ce_sema_t)) ;
 
     if (sem_init(&sp->sema,0,initial_value))
-	ce_panic("hot_sema_Create: sema_init") ;
+	ce_panic("ce_sema_Create: sema_init") ;
     return sp;
 
 }
@@ -229,7 +229,7 @@ ce_sema_Destroy(ce_sema_t *sp) {
     assert(sp) ;
     
     if (sem_destroy(&sp->sema))
-	ce_panic("hot_sema_Destroy");
+	ce_panic("ce_sema_Destroy");
 
     free(sp);
 }
@@ -239,7 +239,7 @@ ce_sema_Inc(ce_sema_t *sp) {
     assert(sp) ;
     TRACE("ce_sema_Inc");
     if (sem_post(&sp->sema))
-	ce_panic("hot_sema_Inc: sem_post") ;
+	ce_panic("ce_sema_Inc: sem_post") ;
 }
 
 void
@@ -260,7 +260,7 @@ ce_lck_Create(void)
     l = (ce_lck_t*) ce_malloc(sizeof(ce_lck_t)) ;
 
     if (pthread_mutex_init(&l->lck,NULL))
-	ce_panic("hot_sema_Create: sema_init") ;
+	ce_panic("ce_sema_Create: sema_init") ;
     return l;
 }
 
@@ -299,11 +299,8 @@ ce_lck_Unlock(ce_lck_t *l) {
 #include <assert.h>
 #include <thread.h>
 #include <synch.h>
-#include "hot_sys.h"
-#include "hot_error.h"
-#include "hot_thread.h"
 
-typedef void*(*hot_thread_routine)(void*);
+typedef void*(*ce_thread_routine)(void*);
 
 /* Create a new thread
  */
@@ -313,7 +310,7 @@ ce_thread_Create(void (*routine)(void*),
 		 int stacksize) 
 {
     if (thr_create(NULL, stacksize, 
-		   (hot_thread_routine) routine, arg, 0, NULL))     
+		   (ce_thread_routine) routine, arg, 0, NULL))     
 	ce_panic("ce_thread_Create: thr_create");
 }
 
@@ -332,7 +329,7 @@ ce_lck_Create(void)
     l = (ce_lck_t*) ce_malloc(sizeof(ce_lck_t)) ;
 
     if (mutex_init(&l->lck, 0, NULL))
-	ce_panic("hot_sema_Create: sema_init") ;
+	ce_panic("ce_sema_Create: sema_init") ;
     return l;
 }
 
@@ -384,7 +381,7 @@ ce_sema_Create(int initial_value)
     sp = (ce_sema_t*) ce_malloc(sizeof(ce_sema_t)) ;
 
     if (sema_init(&sp->sema, initial_value, 0, NULL))
-      ce_panic("hot_sema_Create: sema_init") ;
+      ce_panic("ce_sema_Create: sema_init") ;
     return sp;
 }
 
@@ -396,7 +393,7 @@ ce_sema_Destroy(ce_sema_t *sp)
   if (sema_destroy(&sp->sema))
       ce_panic("ce_sema_Destroy: sema_destroy");
 
-  free(sema);
+  free(sp);
 }
 
 void
@@ -421,3 +418,151 @@ ce_sema_Dec(ce_sema_t *sp)
 
 /**************************************************************/
 /**************************************************************/
+/* The HP-UX, pthreads version
+ */
+#ifdef hpux
+#include <semaphore.h>
+#include <pthread.h>
+
+struct ce_sema {
+    sem_t sema ;
+};
+
+struct ce_lck {
+    pthread_mutex_t lck;
+};
+
+static pthread_t td;
+
+typedef void*(*ce_thread_routine)(void*);
+
+/* Create a new thread
+ */
+void
+ce_thread_Create(void (*routine)(void*), 
+		 void *arg, 
+		 int stacksize)
+{
+  int               attr_stat;
+  /*pthread_attr_t is mainly a pointer on HP-UX!!! */
+  pthread_attr_t    attr;
+    
+  attr_stat = pthread_attr_create(&attr);
+  if (attr_stat == -1 )
+    ce_panic("ce_thread_Create: pthread_attr_create");
+    
+/* hpux threads currently do not support pthread_attr_setstacksize.??
+ */
+#if !(defined(hpux_nostacksize))
+  if (attr_stat != -1)
+  {
+    if (pthread_attr_setstacksize(&attr, stacksize))
+      ce_panic("ce_thread_Create: pthread_attr_setstacksize");
+  }
+#endif
+    
+
+
+#if defined(PTHREAD_CREATE_DETACHED)
+    /* This is done in order to release the thread resources when the
+     * thread dies.  
+     */
+    if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED))
+	ce_panic("ce_thread_Create: pthread_attr_setdetachstate");
+#endif
+
+    
+    if (pthread_create(&td, attr, (ce_thread_routine) routine, arg))
+	ce_panic("ce_thread_Create: pthread_create");
+    pthread_attr_delete(&attr);
+}
+/*endproc ce_thread_Create()*/
+
+
+/* Create a semaphore.
+ */
+ce_sema_t *
+ce_sema_Create(int initial_value)
+{
+    ce_sema_t *sp;
+    
+    sp = (ce_sema_t*) ce_malloc(sizeof(ce_sema_t)) ;
+
+    if (sem_init(&sp->sema,0,initial_value))
+	ce_panic("ce_sema_Create: sema_init") ;
+    return sp;
+
+}
+
+void
+ce_sema_Destroy(ce_sema_t *sp) {
+    assert(sp) ;
+    
+    if (sem_destroy(&sp->sema))
+	ce_panic("ce_sema_Destroy: sem_destroy");
+
+    free(sp);
+}
+
+void
+ce_sema_Inc(ce_sema_t *sp) {
+    assert(sp) ;
+    TRACE("ce_sema_Inc");
+    if (sem_post(&sp->sema))
+	ce_panic("ce_sema_Inc: sem_post") ;
+}
+
+void
+ce_sema_Dec(ce_sema_t *sp) {
+    assert(sp) ;
+    TRACE("ce_sema_Dec");
+    if (sem_wait(&sp->sema))
+	ce_panic("ce_sema_Dec: sem_wait") ;
+}
+
+/* Create a Lock
+ */
+ce_lck_t *
+ce_lck_Create(void)
+{
+    ce_lck_t *l;
+    
+    l = (ce_lck_t*) ce_malloc(sizeof(ce_lck_t)) ;
+
+    if (pthread_mutex_init(&l->lck, pthread_mutexattr_default))
+	ce_panic("ce_lck_Create: pthread_mutex_init") ;
+    return l;
+}
+/*endproc ce_lck_Create()*/
+
+
+void
+ce_lck_Destroy(ce_lck_t *l) {
+    assert(l) ;
+    
+   if (pthread_mutex_destroy(&l->lck))
+	ce_panic("ce_lck_Destroy: pthread_mutex_destroy");
+    free(l);
+}
+/*endproc ce_lck_Destroy()*/
+
+void
+ce_lck_Lock(ce_lck_t *l) {
+    assert(l) ;
+//    TRACE("ce_lck_Lock");
+    if (pthread_mutex_lock(&l->lck))
+	ce_panic("ce_lck_Lock: pthread_mutex_lock") ;
+}
+/*endproc ce_lck_Lock()*/
+
+void
+ce_lck_Unlock(ce_lck_t *l) {
+    assert(l) ;
+//    TRACE("ce_lck_Unlock");
+    if (pthread_mutex_unlock(&l->lck))
+	ce_panic("ce_lck_Unlock: pthread_mutex_unlock") ;
+}
+#endif
+
+/**************************************************************/
+/**************************************************************/
