@@ -3,7 +3,28 @@
 (* Author: Mark Hayden, 8/97 *)
 (**************************************************************)
 
-let unit _ = ()
+type buf = string
+type len = int
+type ofs = int
+type socket = Unix.file_descr
+
+type timeval = {
+  mutable sec10 : int ;
+  mutable usec : int
+} 
+
+type win = 
+    Win_3_11
+  | Win_95_98
+  | Win_NT_3_5
+  | Win_NT_4
+  | Win_2000
+
+type os_t_v = 
+    OS_Unix
+  | OS_Win of win
+
+(**************************************************************)
 
 let print_line s =
   output_string stderr s ;
@@ -21,3 +42,4 @@ let is_unix =
       exit 1
   
 (**************************************************************)
+let max_msg_size = 8 * 1024
